/*--
Copyright (c) 2003  Alexander A. Telyatnikov

Module Name:

    oconsole.c

Abstract:

    Console SQL client to Oracle

Environment:

    Win32.

Notes:


--*/

#include "windows.h"
#include "stdio.h"
#include "conio.h"
#include "ora_common.h"

FILE* f;


/* Define one logon data area and one cursor data area
   Also define a host data area for olog.
   (See ocidfn.h for declarations). */
Lda_Def lda;
Cda_Def cda;
ub4     hda[HDA_SIZE/(sizeof(ub4))];

/* Declare an array of bind values. */
text bind_values[MAX_BINDS][MAX_ITEM_BUFFER_SIZE];

CONSOLE_SCREEN_BUFFER_INFO cons_info;
int output_page_sz = 0;
int saved_output_page_sz = 0;

int print_output = 1;

/* Define arrays of describe and define structs. */
//struct describe desc[MAX_SELECT_LIST_SIZE];
//struct define   def[MAX_SELECT_LIST_SIZE];

/*  Declare this programs functions. */
sword  connect_user();
sword  describe_define();
sword  do_binds();
void   do_exit();
void   oci_error();
sword  get_sql_statement();
void   print_header();

void   print_rows();
void   print_help();

int prompt_login = 1;
text username[132] = "system";
text password[132] = "manager";
text service[132];

/* Globals */
static text sql_statement[2048];
static sword sql_function;
static sword numwidth = 8;

struct describe desc[MAX_SELECT_LIST_SIZE];
struct define   def[MAX_SELECT_LIST_SIZE];

HANDLE StdOut = NULL;
HANDLE StdIn  = NULL;

FILE* fh = NULL; // command file
char fname[1024];

void console_input(char *buf,       // pointer to buf  
                   int  buf_size,   // size of buffer 
                   int  output)     // output flag (1-passwd / 0-nopasswd)
{
    int r = 0; // counter
    int c = 0; // character
    while ( r<buf_size ){
        c = _getch();            
        if (c==8) { // backspase
            if (r) { 
                r--; 
                printf("\b \b");
            }
        }
        else if (c==13) { // Enter
            buf[r] = '\0';
            r++;
            break;
        }
        else {
            buf[r] = (char) c;
            if (output == 0) { _putch(c);  }
            if (output == 1) { _putch('*'); }
            r++;
        }                                        
    } 
}

int
switch_to_file(char* fname) {
    fh = fopen(fname, "r");
    if(!fh) {
        printf("Error: Can't open SQL command file: %s\n", fname);
        return 0;
    }
    saved_output_page_sz = output_page_sz;
    output_page_sz = 0;
    printf("----------------------------\n");
    printf("Start processing SQL command file:\n%s\n", fname);
    return 1;
}

void main (int argc, char* argv[])
{
    sword ncols;
    int i;
//    int j;

    StdOut = GetStdHandle(STD_OUTPUT_HANDLE);
    StdIn  = GetStdHandle(STD_INPUT_HANDLE);
    GetConsoleScreenBufferInfo(StdOut, &cons_info);
    output_page_sz = cons_info.dwSize.Y-1;
    cons_info.dwMaximumWindowSize.X = max(cons_info.dwMaximumWindowSize.X, 800);
    cons_info.dwMaximumWindowSize.Y = max(cons_info.dwMaximumWindowSize.Y, 800);
    SetConsoleScreenBufferSize(StdOut, cons_info.dwMaximumWindowSize);

    printf("Oracle SQL console, version 0.1b\n"
           "Copyright (c) Alexander A. Telyatnikov, 2003,  Home Site: alter.org.ua\n");

/*    username[0] = 0;
    password[0] = 0;*/
    service[0] = 0;

    for(i=1; i<argc; i++) {
        if(!argv[i])
            continue;
        if(!memicmp(argv[i], "user:", 5)) {
            strcpy(&username[0], argv[i]+5);
            prompt_login = 0;
            continue;
        }
        if(!memicmp(argv[i], "pwd:", 4)) {
            strcpy(&password[0], argv[i]+4);
            prompt_login = 0;
            continue;
        }
        if(!memicmp(argv[i], "svc:", 4)) {
            strcpy(&service[0], argv[i]+4);
            prompt_login = 0;
            continue;
        }
        if(!memicmp(argv[i], "service:", 7)) {
            strcpy(&service[0], argv[i]+7);
            prompt_login = 0;
            continue;
        }
        if(argv[i][0] == '@') {
            switch_to_file(argv[i]+1);
            prompt_login = 0;
            continue;
        }
        if(!memicmp(argv[i], "-h", 2) ||
           !memicmp(argv[i], "-?", 2) ||
           !memicmp(argv[i], "-help", 5) ||
           !memicmp(argv[i], "help", 4) ||
           !memicmp(argv[i], "h", 1) ||
           !memicmp(argv[i], "?", 1)) {
            print_help();
            exit(0);
        }
    }
    /* Connect to ORACLE. */
    if (connect_user())
        exit(-1);

    /* Open a cursor, exit on error (unrecoverable). */
    if (oopen(&cda, &lda, (text *) 0, -1, -1, (text *) 0, -1))
    {
        printf("Error opening cursor.  Exiting...\n");
        ologof(&lda);
        exit(-1);
    }

    /* Process user's SQL statements. */
    for (;;)
    {
        /* Get the statement, exit on "exit". */
        if (get_sql_statement()) {
            printf("\n");
            do_exit(0);
        }

        /* Parse the statement; do not defer the parse,
           so that errors come back right away. */
        if (oparse(&cda, (text *) sql_statement, (sb4) -1,
                 (sword) PARSE_NO_DEFER, (ub4) PARSE_V7_LNG))
        {
            oci_error(&cda);
            continue;
        }

        /* Save the SQL function code right after parse. */
        sql_function = cda.ft;

        /* Bind any input variables. */
        if ((ncols = do_binds(&cda, sql_statement)) == -1)
            continue;

        /* If the statement is a query, describe and define
           all select-list items before doing the oexec. */
        if (sql_function == FT_SELECT)
            if ((ncols = describe_define(&cda)) == -1)
                continue;

        /* Execute the statement. */
        if (oexec(&cda))
        {
            oci_error(&cda);
            continue;
        }

        /* Fetch and display the rows for the query. */
        if (sql_function == FT_SELECT)
        {
            print_header(ncols);
            print_rows(&cda, ncols);
        }

        /* Print the rows-processed count. */
        if ((sql_function == FT_SELECT ||
             sql_function == FT_UPDATE ||
             sql_function == FT_DELETE ||
             sql_function == FT_INSERT) &&
             print_output)
            printf("\n%d row%c processed.\n", cda.rpc,
                   cda.rpc == 1 ? ' ' : 's');
        else
            printf("\nStatement processed.\n");

    } /* end for (;;) */

    ologof(&lda);

}     /* end main() */


sword
connect_user()
{
    sword n;

    /* Three tries to connect. */
    for (n = 3; --n >= 0; )
    {
        if(prompt_login) {
            printf("Username: ");
            console_input(&username[0], 131, 0);
            printf("Password: ");
            console_input(&password[0], 131, 1);
            printf("Service: ");
            console_input(&service[0], 131, 0);
        }
//        prompt_login = 1;

        if (olog(&lda, (ub1 *)hda, username, -1, password, -1,
                 (text *) service, -1, (ub4)OCI_LM_DEF))
        {
            printf("Cannot connect as %s.\n", username);
            printf("Try again.\n\n");
        }
        else
        {
            return 0;
        }
    }
    printf("Connection failed.  Exiting...\n");
    return -1;
}


/*  Describe select-list items. */

sword
describe_define(cda)
Cda_Def *cda;
{
    sword col, deflen, deftyp;
    static ub1 *defptr;

    /* Describe the select-list items. */
    for (col = 0; col < MAX_SELECT_LIST_SIZE; col++)
    {
        desc[col].buflen = MAX_ITEM_BUFFER_SIZE;
        if (odescr(cda, col + 1, &desc[col].dbsize,
                   &desc[col].dbtype, &desc[col].buf[0],
                   &desc[col].buflen, &desc[col].dsize,
                   &desc[col].precision, &desc[col].scale,
                   &desc[col].nullok))
        {
            /* Break on end of select list. */
            if (cda->rc == VAR_NOT_IN_LIST)
                break;
            else
            {
                oci_error(cda);
                return -1;
            }
        }
        /* adjust sizes and types for display */
        switch (desc[col].dbtype)
        {
        case NUMBER_TYPE:
            desc[col].dbsize = numwidth;
            /* Handle NUMBER with scale as float. */
            if(desc[col].precision == 0 || desc[col].scale == 0 ||
               (desc[col].scale+desc[col].precision) > 18) {

                defptr = (ub1 *) &def[col].buf;
                deflen = (sword) MAX_ITEM_BUFFER_SIZE ;
                deftyp = SQLT_LNG;
                desc[col].dbtype = SQLT_LNG;

                
            } else
            if (desc[col].scale != 0)
            {
                defptr = (ub1 *) &def[col].flt_buf;
                deflen = (sword) sizeof(float);
                deftyp = FLOAT_TYPE;
                desc[col].dbtype = FLOAT_TYPE;
            }
            else
            {
                defptr = (ub1 *) &def[col].int_buf;
                deflen = (sword) sizeof(sword);
                deftyp = INT_TYPE;
                desc[col].dbtype = INT_TYPE;
            }
            break;
        default:
            if (desc[col].dbtype == DATE_TYPE)
                desc[col].dbsize = 9;
            if (desc[col].dbtype == ROWID_TYPE)
                desc[col].dbsize = 18;
            defptr = def[col].buf;
            deflen = desc[col].dbsize > MAX_ITEM_BUFFER_SIZE ?
              MAX_ITEM_BUFFER_SIZE : desc[col].dbsize + 1;
            deftyp = STRING_TYPE;
            break;
        }
        if (odefin(cda, col + 1,
                   defptr, deflen, deftyp,
                   -1, &def[col].indp, (text *) 0, -1, -1,
                   &def[col].col_retlen,
                   &def[col].col_retcode))
        {
            oci_error(cda);
            return -1;
        }
    }
    return col;
}


/*  Bind input variables. */

sword
do_binds(cda, stmt_buf)
Cda_Def *cda;
text *stmt_buf;
{
    sword i, in_literal, n;
    text *cp, *ph;

    /* Find and bind input variables for placeholders. */
    for (i = 0, in_literal = FALSE, cp = stmt_buf;
              *cp && i < MAX_BINDS; cp++)
    {
        if (*cp == '\'')
            in_literal = ~in_literal;
        if (*cp == ':' && !in_literal)
        {
            for (ph = ++cp, n = 0;
                 *cp && (isalnum(*cp) || *cp == '_')
                     && n < MAX_SQL_IDENTIFIER;
                 cp++, n++
                )
                ;
            *cp = '\0';
            printf("Enter value for %s: ", ph);
            gets((char *) &bind_values[i][0]);

            /* Do the bind, using obndrv().
               NOTE:  the bind variable address must be static.
               This would not work if bind_values were an
               auto on the do_binds stack. */
            if (obndrv(cda, ph, -1, (ub1 *)&bind_values[i][0], -1,
                       VARCHAR2_TYPE,
                       -1, (sb2 *) 0, (text *) 0, -1, -1))
            {
                oci_error(cda);
                return -1;
            }
            i++;
        }   /* end if (*cp == ...) */
    }       /* end for () */
    return i;
}



/*  Clean up and exit.  LDA and CDA are
    global. */
void
do_exit(rv)
sword rv;
{
    if (oclose(&cda))
        fputs("Error closing cursor!\n", stdout);
    if (ologof(&lda))
        fputs("Error logging off!\n", stdout);
    exit(rv);
}


void
oci_error(cda)
Cda_Def *cda;
{
    text msg[512];
    sword n;

    fputs("\n-- ORACLE ERROR --\n", stderr);
    n = oerhms(&lda, cda->rc, msg, (sword) sizeof (msg));
    fprintf(stderr, "%.*s", n, msg);
    fprintf(stderr, "Processing OCI function %s\n",
            oci_func_tab[cda->fc]);
/*    fprintf(stderr, "Do you want to continue? [yn]: ");
    fgets((char *) msg, (int) sizeof (msg), stdin);
    if (*msg != '\n' && *msg != 'y' && *msg != 'Y')
        do_exit(1);*/
    fputc('\n', stdout);
}


int
skip_spaces(char* buf) {
    int i;
    char a;
    for(i=0; a=buf[i]; i++) {
        if(a != ' ' && a != '\t')
            return i;
    }
    return i;
}

void
print_help() {
    printf("Using OConsole from command prompt:\n");
    printf("  oconsole [user:<username>] [pwd:<password>] [[svc|service]:<service name> [@<sql-filename>]\n");
    printf("  defaults:\n");
    printf("    username     system\n");
    printf("    password     manager\n");
    printf("    service      <empty string>\n");
    printf("SQL console comamnd line reference:\n");
    printf("  along with Oracle SQL statements you can use the following commands\n");
    printf("  to control console behavior:\n");
    printf("quit, exit, /q   exit Oracle SQL console\n");
    printf("/page /nopage    enable/disable prompt for scrolling huge SELECT results\n");
    printf("/out  /noout     enable/disable output of SELECT results\n");
    printf("/connect         connect as other user or to other service\n");
    printf("@<filename>      execute SQL statements from file <filename>\n");
    printf("                   recursive calls are not supported\n");
    printf("/help, /h, /?    display this screen\n");
}

sword
get_sql_statement()
{
    text cbuf[1024];
    text *cp;
    sword stmt_level;
    BOOLEAN break_on_q = FALSE;
    BOOLEAN break_name = FALSE;
    LPSTR     lpCmdLine;
    ULONG lim;
    ULONG i;

    for (stmt_level = 1; ;)
    {
        if (stmt_level == 1)
        {
            /* Init statement buffer and print prompt. */
            *sql_statement = '\0';
            fputs("\nOCISQL> ", stdout);
        }
        else
        {
            printf("%3d     ", stmt_level);
        }

        /* Get (part of) a SQL statement. */
        if(!fh) {
            gets(&cbuf[0]);
        } else {
            if(!fgets(&cbuf[0], sizeof(cbuf)-1, fh)) {
                // end
                fclose(fh);
                fh = NULL;
                output_page_sz = saved_output_page_sz;
                printf("\n----------------------------\n");
                printf("%s processed.\n", fname);
                if(!prompt_login)
                    return -1;
                continue;
            } else {
                printf(cbuf);
            }
        }
        if (*cbuf == '\0')
            continue;
        if (strncmp((char *) cbuf, "exit", 4) == 0)
            return -1;
        if (strncmp((char *) cbuf, "quit", 4) == 0)
            return -1;
        if (strncmp((char *) cbuf, "/q", 2) == 0)
            return -1;

        if (*cbuf == '@') {
            if(fh) {
                printf("Error: recursive call of SQL comamnd file is not supported\n");
                continue;
            }
            i = skip_spaces(cbuf+1);
            lpCmdLine = cbuf+i+1;
            lim = strlen(lpCmdLine);
            if(!lim) {
                printf("Error: SQL command file name required\n");
                continue;
            }
            // skip name of executable
            if(lpCmdLine[0] == '\"') {
                // check if executable name comes in quotation marks
                break_on_q = TRUE;
                lpCmdLine++;
            }
            // search for end of executable's name
            for (i=0; !break_name && (i<lim-1) ;i++) {
                switch(lpCmdLine[i]) {
                case '\"':
                    break_name = TRUE;
                    break;
                case '\0':
                case '\t':
                case '\n':
                case '\r':
                    if(!break_on_q) {
                        break_name = TRUE;
                    }
                    break;
                }
            }
            break_name = 0;
            if(i && break_on_q) {
                i--;
            }
            break_on_q = 0;
            if(i<1) {
                printf("Error: SQL command file name required\n");
                continue;
            }
//            strncpy(&fname[0], lpCmdLine, 1023);
            i = min(sizeof(fname)-2, i);
            memcpy(fname, lpCmdLine, i+1);
            fname[i+1] = 0;
            if(!switch_to_file(&fname[0]))
                continue;
/*            fh = fopen(&fname[0], "r");
            if(!fh) {
                printf("Error: Can't open SQL command file: %s\n", fname);
                continue;
            }
            saved_output_page_sz = output_page_sz;
            output_page_sz = 0;
            stmt_level = 1;
            printf("----------------------------\n");
            printf("Start processing SQL command file:\n%s\n", fname);*/
            stmt_level = 1;
            continue;
        }

        if (strncmp((char *) cbuf, "/nopage", 7) == 0) {
            output_page_sz = 0;
            printf("Paging of output is DISABLED\n");
            continue;
        }
        if (strncmp((char *) cbuf, "/page", 5) == 0) {
            GetConsoleScreenBufferInfo(StdOut, &cons_info);
            output_page_sz = cons_info.dwSize.Y-1;
            printf("Paging of output is ENABLED\n");
            continue;
        }

        if (strncmp((char *) cbuf, "/noout", 6) == 0) {
            print_output = 0;
            printf("SQL command output is DISABLED\n");
            continue;
        }
        if (strncmp((char *) cbuf, "/out", 4) == 0) {
            print_output = 1;
            printf("SQL command output is ENABLED\n");
            continue;
        }

        if (strncmp((char *) cbuf, "/help", 4) == 0 ||
            strncmp((char *) cbuf, "/h", 2) == 0 ||
            strncmp((char *) cbuf, "/?", 2) == 0) {
            print_help();
            continue;
        }

        if (strncmp((char *) cbuf, "/connect", 8) == 0) {
            ologof(&lda);
            if (connect_user())
                exit(-1);
            /* Open a cursor, exit on error (unrecoverable). */
            if (oopen(&cda, &lda, (text *) 0, -1, -1, (text *) 0, -1))
            {
                printf("Error opening cursor.  Exiting...\n");
                ologof(&lda);
                exit(-1);
            }
            stmt_level = 1;
            continue;
        }

        /* Concatenate to statement buffer. */
        if (stmt_level > 1)
            strcat((char *) sql_statement, " ");
        strcat((char *) sql_statement, (char *) cbuf);

        /* Check for possible terminator. */
        cp = &sql_statement[strlen((char *) sql_statement) - 1];

        while (isspace(*cp))
            cp--;
        if (*cp == ';')
        {
            *cp = '\0';
            break;
        }
        stmt_level++;
    }
    return 0;
}


void
print_header(ncols)
sword ncols;
{
    sword col, n;

    if(!print_output)
        return;

    fputc('\n', stdout);

    for (col = 0; col < ncols; col++)
    {
        n = desc[col].dbsize - desc[col].buflen;
        if (desc[col].dbtype == FLOAT_TYPE ||
            desc[col].dbtype == INT_TYPE)
        {
            printf("%*c", n, ' ');
            printf("%*.*s", desc[col].buflen,
                   desc[col].buflen, desc[col].buf);
        }
        else
        {
            printf("%*.*s", desc[col].buflen,
                   desc[col].buflen, desc[col].buf);
            printf("%*c", n, ' ');
        }
        fputc(' ', stdout);
    }
    fputc('\n', stdout);

    for (col = 0; col < ncols; col++)
    {
        for (n = desc[col].dbsize; --n >= 0; )
            fputc('-', stdout);
        fputc(' ', stdout);
    }
    fputc('\n', stdout);
}


void
print_rows(Cda_Def *cda,
           sword ncols)
{
    sword col, n;
    int row;
    int key = 0;
    ULONG tmp;
    INPUT_RECORD inp;

    if(!print_output)
        return;
    if(output_page_sz) {
        GetConsoleScreenBufferInfo(StdOut, &cons_info);
        output_page_sz = cons_info.srWindow.Bottom-cons_info.srWindow.Top;
    }

    for (row=0;;row++)
    {
        fputc('\n', stdout);
        if(output_page_sz && row >= output_page_sz) {
            printf("Press any key to continue or Esc to cancel");
//            tmp = GetConsoleMode(StdIn, &mode_orig);
//            tmp = SetConsoleMode(StdIn, mode_orig & ~ENABLE_LINE_INPUT);
            while(1) {
                key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
                if(key && inp.EventType == KEY_EVENT) {
                    if(inp.Event.KeyEvent.bKeyDown &&
                        inp.Event.KeyEvent.wVirtualScanCode == 0x01 /*VK_ESC*/) {
                        fputc('\n', stdout);
                        return;
                    }
                    if(inp.Event.KeyEvent.bKeyDown) {
                        fputc('\n', stdout);
                        row=0;
                        break;
                    }
                }
//                key = ReadConsole(StdIn, &key, 1, &tmp, NULL);
                Sleep(50);
            }
//            SetConsoleMode(StdIn, mode_orig);
        }

        Sleep(1);
        while(GetNumberOfConsoleInputEvents(StdIn, &key) && key) {
//        while(WaitForSingleObject(StdIn, 0) == WAIT_OBJECT_0) {
            key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
            if(key && inp.EventType == KEY_EVENT) {
                if(inp.Event.KeyEvent.bKeyDown &&
                    inp.Event.KeyEvent.wVirtualScanCode == 0x01 /*VK_ESC*/) {
                    fputc('\n', stdout);
                    return;
                }
            }
        }
        /* Fetch a row.  Break on end of fetch,
           disregard null fetch "error". */
        for (col = 0; col < ncols ; col++) {
            memset(&def[col], 0, sizeof(def[col]));
        }
        if (ofetch(cda))
        {
            if (cda->rc == NO_DATA_FOUND)
                break;
            if (cda->rc != NULL_VALUE_RETURNED)
                oci_error(cda);
        }
        for (col = 0; col < ncols ; col++)
        {
            /* Check col. return code for null.  If
               null, print n spaces, else print value. */
            if (def[col].indp < 0) {
                printf("%*c", desc[col].dbsize, ' ');
            } else {
                switch (desc[col].dbtype)
                {
                case FLOAT_TYPE:
                    printf("%*.*f", numwidth, 2, def[col].flt_buf);
                    break;
                case INT_TYPE:
                    printf("%*d", numwidth, def[col].int_buf);
                    break;
                case SQLT_LNG:
                    printf("%s", def[col].buf);
                    break;
                default:
                    printf("%s", def[col].buf);
                    n = desc[col].dbsize - strlen((char *) def[col].buf);
                    if (n > 0)
                        printf("%*c", n, ' ');
                    break;
                }
            }
            fputc('\t', stdout);
        }
    }  /* end for (;;) */
}

